/*
	Author:  Jamie A. Grant
   Date:    September 17, 2004
   Version: 1.0
   Courtesy of AV-Base Systems Inc.
   Licence: Free use for the dBase community, including commercial use.
            Only request is that this header be maintained.
            Code may be modified or used as desired.

	Required File:
   			IAssocArray.cc

	Example:	set procedure to XMLParser.cc additive
            xml = new XMLParser(strFile)
            o = xml.Parse()		//Returns XMLObject matching the XML structures.
*/

class XMLParser( strNewFile )

   //Internal Properties
	protect this.strFile
   this.strFile = NULL

   protect this.aTokens
   this.aTokens = new array()

   //Internal Functions
   protect AddAttributes
   protect FormatAttribute
   protect GetAttributes
   protect GetTokenName
   protect IsValidAttribute
   protect IsValidTag
   protect ReplaceSpecialChars

   //Initialize Object
   if NOT empty(strNewFile)
   	class::SetFile( strNewFile )
   endif

	function Reset

   	this.strFile	:= NULL
      this.aTokens	:= new array()

   return

   function GetFile

   return this.strFile

   function SetFile( strFile )

   	local bValid
      bValid = false

      TRY
      	bValid = file(strFile)
      CATCH( Exception e )
      ENDTRY
      if bValid
   		this.strFile := strFile
      endif

   return bValid

   function Tokenize
   	//Note:	1) chr(13) + chr(10) are retained at the end of each line inside XML Comments and CData.
      //			2) Special XML characters are converted, per the ReplaceSpecialChars function.
      //			3) Invalid tokens are ignored. This includes:
      //					a) XML Comments or CData tags that are not closed.
      //					b) Tags with opening < that do not have closing >
      //			4) Attributes (values contained within a tag) are converted to separate sub-tags.
      //			5) XML namespaces are not handled.

   	local aArray
      aArray = new array()

      //Read in the entire file, break each word/token
      //into consecutive items in the array.
		local f, strCurrent, strLength, i, strNew, chrCurrent, bComments, bCData, aAttr
      strNew		= ""
      bComments	= false
      bCData		= false
      f = new file()
      TRY
      	f.open( this.strFile )
      CATCH( Exception e )
      	//File may have been deleted in the meantime.
      	msgbox("Error opening file: " + this.strFile, "Alert", 48)
         return new array()
      ENDTRY

      do while not f.eof()
         strCurrent	= f.gets() + chr(13) + chr(10)
         strLength	= len(strCurrent)

         for i = 1 to strLength
         	chrCurrent = strCurrent.charAt( i-1 )	//charAt is 0-based.
            if (chrCurrent == chr(13) OR chrCurrent == chr(10)) AND NOT bComments AND NOT bCData
            	//Line return characters are avoided.
               //Except as part of XML comments or CData, in which case they are included.
            elseif chrCurrent == [<]
               if bComments OR bCData
               	strNew += chrCurrent
               else
                  if NOT empty(strNew)
                     if class::IsValidTag(strNew)
                     	aArray.add(strNew)
                        aAttr = class::GetAttributes(strNew)
                        if aAttr.count() > 0
                        	class::AddAttributes(aArray, aAttr)
                        endif
                     endif
                  endif
                  strNew = "<"
               endif
               if NOT (i+2) > strLength AND (strCurrent.charAt(i) == [!] AND strCurrent.charAt(i+1) == [-] AND strCurrent.charAt(i+2) == [-])
               	bComments = true
               elseif NOT (i+8) > strLength AND substr(strCurrent, i, 9) == '<![CDATA['
               	bCData = true
               endif
            elseif chrCurrent == [>]
               if bComments 
               	if len(strNew) > 2 AND right(strNew, 2) == [--]
                     //Commented section of XML, comments end with -->
                     bComments = false
                     strNew += [>]
                     aArray.add(strNew)
                     aAttr = class::GetAttributes(strNew)
                     if aAttr.count() > 0
                        class::AddAttributes(aArray, aAttr)
                     endif
                  	strNew = ""
                  else
                  	strNew += chrCurrent
                  endif
               elseif bCData
               	if len(strNew) > 2 AND right(strNew, 2) == ']]'
                     //CData section of XML, ends with ]]>
                     bCData = false
                     strNew += [>]
                     aArray.add(strNew)
                     aAttr = class::GetAttributes(strNew)
                     if aAttr.count() > 0
                        class::AddAttributes(aArray, aAttr)
                     endif
                  	strNew = ""
                  else
                  	strNew += chrCurrent
                  endif
               else
               	strNew += [>]
                  aArray.add(strNew)
                  aAttr = class::GetAttributes(strNew)
                  if aAttr.count() > 0
                     class::AddAttributes(aArray, aAttr)
                  endif
                  strNew = ""
               endif
            else
            	strNew += chrCurrent
            endif
         next
		enddo
      f.close()
      f = NULL

      //Add the final token.
      if NOT empty(strNew)
      	if bComments AND NOT bCData
         	//Both sections must have the expected closing tags.
         elseif NOT class::IsValidTag( strNew )
         	//Opening tag does close properly.
      	else
         	aArray.add(strNew)
            aAttr = class::GetAttributes(strNew)
            if aAttr.count() > 0
               class::AddAttributes(aArray, aAttr)
            endif
         endif
      endif

      aArray = class::ReplaceSpecialChars( aArray )

   return aArray

   function GetTokenName( strPassed )

   	local strCurrent
      strCurrent = trim(ltrim(strPassed))

      //Take off the open/close brackets.
      if len(strCurrent) > 0 AND left(strCurrent, 1) == [<]
      	strCurrent = right(strCurrent, len(strCurrent) - 1)
      endif
      if len(strCurrent) > 0 AND left(strCurrent, 1) == [/]
      	strCurrent = right(strCurrent, len(strCurrent) - 1)
      endif
      if len(strCurrent) > 1 AND right(strCurrent, 2) == [/>]
      	strCurrent = left(strCurrent, len(strCurrent) - 2)
      elseif len(strCurrent) > 0 AND right(strCurrent, 1) == [>]
      	strCurrent = left(strCurrent, len(strCurrent) - 1)
      endif
      strCurrent = trim(ltrim(strCurrent))

      //The name includes everything up the first space or the end of the string.
      local idx, idxSpace, idxTab, strName
      strName = ""
      idxSpace = at( [ ], strCurrent )
      idxTab = at( chr(9), strCurrent )
      if idxSpace == 0 AND idxTab == 0
      	strName = strCurrent
      elseif idxSpace == 0 AND NOT idxTab == 0
			strName = left(strCurrent, idxTab - 1)
      elseif NOT idxSpace == 0 AND idxTab == 0
        	strName = left(strCurrent, idxSpace - 1)
		else
         idx = iif( idxSpace < idxTab, idxSpace, idxTab )
         strName = left(strCurrent, idx - 1)
      endif

   return strName

   function ReplaceSpecialChars( aArray )
   	//XML standard does not permit certain characters.
      //They are represented by certain codes instead.
      //Not done for XML Comments or CData.

   	local i, str
      for i = 1 to aArray.size
      	str = "" + aArray[i]
         if len(str) > 3 AND left(str, 4) == [<!--]
         	//XML comments can have the invalid characters.
            LOOP
         elseif len(str) > 8 AND left(str, 9) == '<![CDATA['
         	//XML CData can have the invalid characters.
            LOOP
         endif

      	do while [&apos;] $ str
            str = stuff(str, at( [&apos;], str ), 6, ['])
         enddo
         do while [&quot;] $ str
            str = stuff(str, at( [&quot;], str ), 6, ["])
         enddo
         do while [&amp;] $ str
            str = stuff(str, at( [&amp;], str ), 5, [&])
         enddo
         do while [&gt;] $ str
            str = stuff(str, at( [&gt;], str ), 4, [>])
         enddo
         do while [&lt;] $ str
            str = stuff(str, at( [&lt;], str ), 4, [<])
         enddo

         aArray[i] = str
      next

   return aArray

   function IsValidTag( strPassed )

      local bValid
   	if len(strPassed) > 0 AND left(strPassed, 1) == [<] AND NOT right(strPassed,1) == [>]
      	bValid = false
      else
      	bValid = true
      endif

   return bValid

   function IsValidAttribute( strPassed )

   	local bValid
      bValid = false

      if len(strPassed) > 1
      	if left(strPassed, 1) == ["] AND right(strPassed, 1) == ["]
         	//Start and ends with double quotes, no double quotes inside.
            if NOT ["] $ substr(strPassed, 2, len(strPassed) - 2)
            	bValid = true
            endif
         elseif left(strPassed, 1) == ['] AND right(strPassed, 1) == [']
         	//Start and ends with single quotes, no single quotes inside.
            if NOT ['] $ substr(strPassed, 2, len(strPassed) - 2)
            	bValid = true
            endif
         endif
      endif

   return bValid

   function FormatAttribute( strPassed )

   	local str
      str = ""
      if len(strPassed) > 1
      	str = substr(strPassed, 2, len(strPassed) - 2)
      endif

		do while [&apos;] $ str
         str = stuff(str, at( [&apos;], str ), 6, ['])
      enddo
      do while [&quot;] $ str
         str = stuff(str, at( [&quot;], str ), 6, ["])
      enddo
      do while [&amp;] $ str
         str = stuff(str, at( [&amp;], str ), 5, [&])
      enddo
      do while [&gt;] $ str
         str = stuff(str, at( [&gt;], str ), 4, [>])
      enddo
      do while [&lt;] $ str
         str = stuff(str, at( [&lt;], str ), 4, [<])
      enddo

   return str

   function GetAttributes( strPassed )
   	//Similar to Tokenize, but intended for attributes inside an XML tag.

      set procedure to IAssocArray.cc additive
   	local aArray
      aArray = new IAssocArray()

		if len(strPassed) > 1 AND left(strPassed, 2) == [<?]
      	return aArray
      elseif len(strPassed) > 1 AND left(strPassed, 2) == [<!]
      	return aArray
      elseif len(strPassed) > 8 AND left(strPassed, 9) == "<![CDATA["
      	return aArray
      endif

      local strCurrent
      strCurrent = trim(strPassed)
      if len(strCurrent) > 1 AND right(strCurrent, 2) == [/>]
      	strCurrent = left(strCurrent, len(strCurrent) - 2)
      elseif len(strCurrent) > 0 AND right(strCurrent, 1) == [>]
      	strCurrent = left(strCurrent, len(strCurrent) - 1)
      endif

      local strLength, strTokenName, idx, idxStart, i, chrCurrent, strNew, strTag
      strLength		= len(strCurrent)
      strTokenName	= class::GetTokenName( strCurrent )
      idx				= at( strTokenName, strCurrent )
      idxStart			= idx + len(strTokenName)
      strNew			= ""
      strTag			= ""

		for i = idxStart to strLength
         chrCurrent = strCurrent.charAt( i-1 )	//charAt is 0-based.
         if chrCurrent == [ ] OR chrCurrent == chr(9)		//Space or Tab
            if NOT empty(strNew)
            	if empty(strTag)
               	strTag = strNew
               else
                  if NOT aArray.isKey(strTag)	//Ignore duplicate attributes.
                     if class::IsValidAttribute(strNew)
                     	strNew = class::FormatAttribute( strNew )
            				aArray.AddItem( strTag, strNew )
                     endif
                     strTag = ""
                  endif
               endif
               strNew = ""
            endif
         elseif chrCurrent == [=]
            if NOT empty(strNew)
            	strTag = strNew
            endif
            strNew = ""
         else
         	strNew += chrCurrent
         endif
      next

      if NOT empty(strTag) AND NOT empty(strNew)
      	if NOT aArray.isKey(strTag)
            if class::IsValidAttribute(strNew)
               strNew = class::FormatAttribute( strNew )
					aArray.AddItem( strTag, strNew )
            endif
         endif
      endif

   return aArray

   function AddAttributes(aArray, aAttr)

      if empty(aArray) AND empty(aAttr)
      	msgbox("AddAttributes", "JAG")
      	return
      endif

      local i, aSize
      aSize = aAttr.count()
      if aSize > 0
         local strKey
         strKey = aAttr.GetFirstKey()

         for i = 1 to aSize
            aArray.add( [<] + strKey + [>] )
            aArray.add( aAttr[strKey] )
            aArray.add( [</] + strKey + [>] )
            strKey = aAttr.GetNextKey(strKey)
         next i
      endif

   return

   function Parse
   	//Builds XMLObject hierachy according to the XML structure.
      //Excludes: XML Comments <!--
      //				CData <![CDATA[
      //				Special XML Data <?
      //Ignores:	Doesn't care if closing tag is missing.
      //				Doesn't care if closing tag is there with no opening tag before it.

   	local oXML, intLevel
      oXML		= new XMLObject()
      intLevel	= 0
      
      if empty(this.strFile)
      	return oXML
      endif
      this.aTokens := class::Tokenize()

      //Private vars so that dBase Debugger can see them.
      private i, strTokenName, strTokenLen, strCloseName, aCurrent, strValue, aSwap, aNewAssoc, strSwap, oNew, aNewArray
      strTokenName	= NULL
      strValue			= ""
      oCurrent			= oXML
      for i = 1 to this.aTokens.size
      	strTokenLen = len(this.aTokens[i])
         if strTokenLen > 1 AND left(this.aTokens[i], 2) == [<?]
				//Could add logic to check "xml version...encoding..." later on.
         	LOOP
         elseif strTokenLen > 3 AND left(this.aTokens[i], 4) == [<!--]
         	//Exclude all comments.
            LOOP
         elseif strTokenLen > 8 AND left(this.aTokens[i], 9) == '<![CDATA['
         	//Exclude all CData sections.
            LOOP
         endif

         if strTokenLen > 1 AND left(this.aTokens[i], 2) == [</]
         	//Close Token
         	strCloseName = class::GetTokenName( this.aTokens[i] )

            if strCloseName == strTokenName
            	oNew = new XMLString()
               oNew.parent		:= oCurrent
               oNew.ParentKey	:= strTokenName
               oNew.String		:= strValue
               oNew.Level		:= intLevel

               if NOT oCurrent.value.isKey(strTokenName)
               	oCurrent.value.AddItem( strTokenName, oNew )
               else
               	if upper(oCurrent.value[strTokenName].className) == "ARRAY"
                  	oCurrent.value[strTokenName].add( oNew )
                  else
                     oSwap = oCurrent.value[strTokenName]
                     aNewArray = new XMLArray()
                     aNewArray.Parent		:= oCurrent
                     aNewArray.ParentKey	:= strTokenName
                     aNewArray.Level		:= intLevel
                     oSwap.Parent			:= aNewArray
                     oSwap.ParentKey		:= 1
                     oNew.Parent				:= aNewArray
                     oNew.ParentKey			:= 2
                     aNewArray.add( oSwap )
                     aNewArray.add( oNew )
                     oCurrent.value[strTokenName] = aNewArray
                  endif
               endif
               strTokenName = NULL
            else
               intLevel--
               if oCurrent.ParentKey == strCloseName
            		oCurrent	= oCurrent.parent
               else
               	oCurrent	= oCurrent.parent
                  if oCurrent.ParentKey == strCloseName
                     oCurrent	= oCurrent.parent
                  endif
               endif
            endif
            strValue = ""
         elseif strTokenLen > 0 AND left(this.aTokens[i], 1) == [<]
         	//Open Token, must come after Close Token check.
            if NOT empty(strTokenName)
               oNew = new XMLObject()
               oNew.parent		:= oCurrent
               oNew.ParentKey	:= strTokenName
               oNew.Level		:= intLevel

            	if NOT oCurrent.value.isKey(strTokenName)
                  oCurrent.value.AddItem( strTokenName, oNew )
               else
                  if upper(oCurrent.value[strTokenName].className) == "XMLOBJECT"
                     oSwap = oCurrent.value[strTokenName]
                     aNewArray = new XMLArray()
                     aNewArray.Parent		:= oCurrent
                     aNewArray.ParentKey	:= strTokenName
                     aNewArray.Level		:= intLevel
                     oSwap.Parent			:= aNewArray
                     oSwap.ParentKey		:= 1
                     oNew.Parent				:= aNewArray
                     oNew.ParentKey			:= 2
                     aNewArray.add( oSwap )
                     aNewArray.add( oNew )
                     oCurrent.value[strTokenName] = aNewArray
                  else
                     oNew.Parent		:= oCurrent.value[strTokenName]
                     oNew.ParentKey	:= oCurrent.value[strTokenName].size + 1
                  	oCurrent.value[strTokenName].add( oNew )
                  endif
               endif
               intLevel++
               oCurrent = oNew
            endif

         	strTokenName	= class::GetTokenName( this.aTokens[i] )
            strValue			= ""
            if strTokenLen > 1 AND right(this.aTokens[i], 2) == [/>]
            	//Empty tag, no associated value.
               oNew = new XMLString()
               oNew.parent		:= oCurrent
               oNew.ParentKey	:= strTokenName
               oNew.String		:= strValue
               oNew.Level		:= intLevel 
               if NOT oCurrent.value.isKey(strTokenName)
               	oCurrent.value.AddItem( strTokenName, oNew )
               else
               	if upper(oCurrent.value[strTokenName].className) == "ARRAY"
                     oNew.Parent		:= oCurrent.value[strTokenName]
                     oNew.ParentKey	:= oCurrent.value[strTokenName].size + 1
                  	oCurrent.value[strTokenName].add( oNew )
                  else
                     oSwap = oCurrent.value[strTokenName]
                     aNewArray = new XMLArray()
                     aNewArray.Parent		:= oCurrent
                     aNewArray.ParentKey	:= strTokenName
                     aNewArray.Level		:= intLevel
                     oSwap.Parent			:= aNewArray
                     oSwap.ParentKey		:= 1
                     oNew.Parent				:= aNewArray
                     oNew.ParentKey			:= 2
                     aNewArray.add( oSwap )
                     aNewArray.add( oNew )
                     oCurrent.value[strTokenName] = aNewArray
                  endif
               endif
               strTokenName = NULL
            endif
         elseif NOT empty(strTokenName)
         	//Excludes spaces that are inserted between the close of one token and the opening of another.
            //Includes spaces that are included in the value between an open and a close token.
            strValue = this.aTokens[i]
         endif
      next

   return oXML

   function GetNextObject( oPassed )

   	local oNext, oCurrent, strKey, bGetOut
      oNext		= NULL
      oCurrent = oPassed
      bGetOut	= false

      do while NOT oCurrent == NULL
         if oCurrent.ClassName == "XMLOBJECT"
            oNext = oCurrent.value[ oCurrent.value.GetFirstKey() ]
            EXIT
         elseif oCurrent.ClassName == "XMLARRAY"
         	oNext = oCurrent[1]
            EXIT
         else
            if oCurrent.Parent == NULL
            	EXIT
            else
               do while NOT oCurrent == NULL
                  if oCurrent.Parent == NULL
                     bGetOut = true
            			EXIT
               	elseif oCurrent.parent.ClassName == "XMLOBJECT"
                     strKey	= oCurrent.ParentKey
                     oCurrent = oCurrent.parent

                     local nextKey
                     nextKey = oCurrent.value.GetNextKey( strKey )
                     if NOT nextKey == NULL
                        oNext	= oCurrent.value[nextKey]
                        bGetOut = true
                        EXIT
                     endif
                  elseif oCurrent.parent.ClassName == "XMLARRAY"
                  	strKey	= oCurrent.ParentKey
                     oCurrent = oCurrent.parent

                     if oCurrent.size > strKey
                     	oNext	= oCurrent[strKey+1]
                        bGetOut = true
                        EXIT
                     endif
                  else
                  	oCurrent = oCurrent.parent
                  endif
               enddo
            endif
         endif

         if bGetOut == true
         	EXIT
         endif
      enddo

   return oNext

   function SaveFile( strFile, oPassed )
   	//Assumes file has been created already.

      local f
   	f = new file()
      TRY
      	f.open( strFile, "RW" )
      CATCH( Exception e )
      	//File may have been deleted in the meantime.
      	msgbox("Error opening file: " + strFile, "Alert", 48)
         return false
      ENDTRY

   	local oCurrent, strKey
      oCurrent = oPassed

      do while NOT oCurrent == NULL
         if oCurrent.ClassName == "XMLOBJECT"
         	if NOT oCurrent.parent == NULL AND NOT oCurrent.parent.ClassName == "XMLARRAY"
               strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
               strOut = "<" + oCurrent.ParentKey + ">"
               f.puts(strTabs + strOut)
            endif
            oCurrent = oCurrent.value[ oCurrent.value.GetFirstKey() ]
            LOOP
         elseif oCurrent.ClassName == "XMLARRAY"
            if NOT oCurrent.parent == NULL AND NOT oCurrent.parent.ClassName == "XMLARRAY"
               strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
               strOut = "<" + oCurrent.ParentKey + ">"
               f.puts(strTabs + strOut)
            endif
         	oCurrent = oCurrent[1]
            LOOP
         else		//ClassName == XMLSTRING
            strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
            strOut = "<" + oCurrent.ParentKey + ">" + oCurrent + "</" + oCurrent.ParentKey + ">"
            f.puts(strTabs + strOut)

            do while NOT oCurrent == NULL
               if oCurrent.Parent == NULL
                  oCurrent = oCurrent.Parent
                  EXIT
               elseif oCurrent.parent.ClassName == "XMLOBJECT"
                  strKey	= oCurrent.ParentKey
                  oCurrent = oCurrent.parent

                  local nextKey
                  nextKey = oCurrent.value.GetNextKey( strKey )
                  if NOT nextKey == NULL
                     oCurrent	= oCurrent.value[nextKey]
                     EXIT
                  else
                     if NOT oCurrent.parent == NULL AND NOT oCurrent.parent.ClassName == "XMLARRAY"
                        strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
                        strOut = "</" + oCurrent.ParentKey + ">"
                        f.puts(strTabs + strOut)
                     endif
                  endif
               elseif oCurrent.parent.ClassName == "XMLARRAY"
                  strKey	= oCurrent.ParentKey
                  oCurrent = oCurrent.parent

                  if oCurrent.size > strKey
                  	strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
                     strOut = "</" + oCurrent.ParentKey + ">"
                     f.puts(strTabs + strOut)

                     strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
                     strOut = "<" + oCurrent.ParentKey + ">"
                     f.puts(strTabs + strOut)

                     oCurrent	= oCurrent[strKey+1]
                     EXIT
                  else
                  	strTabs = iif( NOT oCurrent.level > 0, "", Replicate( chr(9), oCurrent.level ) )
                     strOut = "</" + oCurrent.ParentKey + ">"
                     f.puts(strTabs + strOut)
                  endif
               endif
            enddo
         endif
      enddo

      f.close()
      f = NULL

	return true
endclass

class XMLObject

   set procedure to IAssocArray.cc additive
   this.ClassName = "XMLOBJECT"
   this.parent		= NULL
   this.parentKey	= NULL
   this.Level		= 0
	this.value		= new IAssocArray()

endclass

class XMLString of String

   this.parent		= NULL
   this.parentKey	= NULL
   this.Level		= 0

endclass

class XMLArray of Array

   this.parent		= NULL
   this.parentKey	= NULL
   this.Level		= 0

endclass
